/*

			     Bullets  --  The Screen Saver
			
		     by John Walker  --  kelvin@fourmilab.ch
		
					Release 1.0  --  May 1996
					Release 2.0  --  July 2006

*/ 

#include "bullets.h"      

#define ID_TIMER 200

/* Function prototypes */

static void GetIniEntries(void);
static void GetIniSettings(void);

/* Global used by SCRNSAVE.LIB. Required for all screen savers. */

char szAppName[40];

/* Globals exported. */

char szName[] = "Bullets Screen Saver"; 

/* Externals defined in SCRNSAVE.LIB. Required for all screen savers. */

HINSTANCE hMainInstance;
HWND hMainWindow;
char szName[TITLEBARNAMELEN];
char szScreenSaver[22];
char szHelpFile[MAXFILELEN];
char szNoHelpMemory[BUFFLEN];
UINT MyHelpMessage;
HOOKPROC fpMessageFilter;

//	Configuration parameters

DWORD resetInterval = 0;					// Reset interval in ticks, 0 = never

static WORD wElapse = 500;					// Speed parameter
static WORD wTimer = 0;						// Timer id
static char szGunfire[] = "Gunfire";		// Gunfire ?
BOOL soundGunfire = TRUE;					// Gunfire sounds ?
static char szBmpfile[] = "BMPfile";		// Bitmap file name
static char bmpfile[MAX_PATH] = "";     	// Bitmap file

static BOOL noMem = FALSE;					// Insufficient memory detected
static char szBmpshow[] = "BMPshow";		// Show BMP file ?
static BOOL bmpShow = FALSE;				// Show BMP file ?
static BOOL bmpLoaded = FALSE;				// BMP file loaded
static RECT bmpRect;						// Rectangle where bitmap loaded

static int burstN = 0;						// Nonzero if doing burst
static BOOL needErase = TRUE;				// Need to erase and repaint screen ?

//  RSTRING  --  Retrieve a string from the resource file.

char *rstring(int resid)
{
#define maxCStrings 10              // Maximum concurrently used strings
    static char rstrings[maxCStrings][80];
    static int n = 0;
    int m = n;

    if (LoadString(hInst, resid, rstrings[m], 79) < 0) {
        strcpy(rstrings[m], "");
    }
    n = (n + 1) % maxCStrings;
    return rstrings[m];
}

/*	RFILTER  --  Load a filter from the resource file and convert
				 vertical bars to zero characters.  */
				 
char *rfilter(int resid)
{
	char *cp, *cp1;
	
	cp1 = cp = rstring(resid);
	while (*cp1 != 0) {
		if (*cp1 == '|') {
			*cp1 = 0;
		}
		cp1++;
	}
	return cp;
}				

/* ScreenSaverProc -- Main entry point for screen saver messages.
					  This function is required for all screen savers.

	Params:  Standard window message handler parameters.

	Return:  The return value depends on the message.

	Note that all messages go to the DefScreenSaverProc(), except
	for ones we process.
*/

WindowProc(ScreenSaverProc)
{
    HDC hDC;
    static RECT rc;
	PAINTSTRUCT ps;

    switch (Message) {

        case WM_CREATE:
        {
            /* Load the strings from the STRINGTABLE */

            GetIniEntries();
            GetIniSettings();            
            if (!saverInit(hWnd)) {
            	noMem = TRUE;
            }

            /* Create a timer to update the image */

            wTimer = SetTimer(hWnd, ID_TIMER, 250, NULL);
            break;
        }

        case WM_TIMER:
			InvalidateRect(hWnd, NULL, FALSE);
            break;

        case WM_DESTROY:
            if (wTimer) {
            	KillTimer(hWnd, ID_TIMER);
            }
            saverTerm();
            break;
        
        case WM_PAINT:
            memset(&ps, 0, sizeof(PAINTSTRUCT));
            hDC = BeginPaint(hWnd, &ps);
            GetClientRect(hWnd, &rc);
			if (needErase) {
				if (bmpfile[0] != 0 && bmpShow) {
					HFILE bmp = _lopen(bmpfile, OF_READ);

					if (bmp != HFILE_ERROR) {
						LONG bmpl = _llseek(bmp, 0L, FILE_END);
						LPBYTE bumpy = (LPBYTE) GlobalAlloc(GPTR, bmpl);

						if (bumpy != NULL) {
							_llseek(bmp, 0L, FILE_BEGIN);
							_hread(bmp, bumpy, bmpl);
							if (memcmp(bumpy, "BM", 2) == 0) {
	    						GetClientRect(hWnd, &rc);
	    						FillRect(hDC, &rc, (HBRUSH) GetStockObject(BLACK_BRUSH));
								bmpLoaded = paintDIB(hWnd, hDC, bumpy, &bmpRect);
							}
							GlobalFree(bumpy);
						}
						_lclose(bmp);
					}
				} else if (fChildPreview) {
				
					/* If we're in preview mode and not displaying a fixed
					   bitmap background, scale the current desktop into the
					   little toy screen in the configuration applet. */
					   
					HWND desk = GetDesktopWindow();
					RECT dr;
					HDC deskDC;

					GetClientRect(desk, &dr);
					deskDC = GetDC(desk);
	    			GetClientRect(hWnd, &rc);
					SetStretchBltMode(hDC, STRETCH_DELETESCANS);
					StretchBlt(hDC, rc.left, rc.top, rc.right - rc.left,
							   rc.bottom - rc.top, deskDC, dr.left, dr.top,
							   dr.right - dr.left, dr.bottom - dr.top, SRCCOPY);
					ReleaseDC(desk, deskDC);
				}
				needErase = FALSE;
			}
			if (!noMem) {
				static int bmult = FALSE,
						   weapShot;

				if (burstN == 0) {
					weapShot = random() % ARSENAL;
					if ((random() % 100) < weapons[weapShot].burstprob) {
						burstN = 2 + (((random() >> 16) *
								 (weapons[weapShot].burstmax - 2)) / 32767);
						bmult = FALSE;
					}
				}
				if (saverMake(hWnd, hDC, weapShot, bmult,
							  bmpLoaded ? bmpRect : rc)) {
					if (burstN > 0) {
						burstN--;
						bmult = burstN > 0;
					}						
					SetTimer(hWnd, ID_TIMER, bmult ?
							 (1000 / weapons[weapShot].reprate)  :
							 (500 + (3000 * (random() >> 16)) / 32767), NULL);
				} else {
					noMem = TRUE;
				}
			}
			
			/*  If we've run out of memory (whether during initialisation or
				during generation), switch the timer to a ten second interval
				and display a randomly-positioned out of memory message on the
				screen.  */
				 
			if (noMem) {
				char s[132];
				int tx, ty;
				SIZE sz;
				
				if (wElapse < 10000) {
					// First time.  Change timer interval to 10 seconds
					if (wTimer) {
						KillTimer(hWnd, ID_TIMER);
					}
					wElapse = 10000;
					wTimer = SetTimer(hWnd, ID_TIMER, wElapse, NULL);
					saverTerm();			// Release all memory in use
				}
	    		GetClientRect(hWnd, &rc);
	    		FillRect(hDC, &rc, (HBRUSH) GetStockObject(BLACK_BRUSH));
  				LoadString(hMainInstance, idsInsuffMemory, s, sizeof s);
	  			GetTextExtentPoint32(hDC, s, lstrlen(s), &sz);
  				SetBkMode(hDC, TRANSPARENT);
  				tx = (int) (((rc.right - sz.cx) * (random() & 32767L)) / 32767L); 
  				ty = (int) (((rc.bottom - sz.cy) * (random() & 32767L)) / 32767L); 
  				SetTextColor(hDC, RGB(255, 255, 255));
  				SetTextAlign(hDC, TA_NOUPDATECP);
  				TextOut(hDC, tx, ty, s, lstrlen(s));
				}
            EndPaint(hWnd, &ps);
            break;

		case WM_ERASEBKGND:
			//	Need to catch this message to bypass default screen clear
			needErase = TRUE;
			return TRUE;

#ifdef CLIPPO

		/*	If CLIPPO is defined, right mouse button copies screen
			to the clipboard.  Why?  So we can make screen shots to
			illustrate the program.  */

		case WM_RBUTTONDOWN:
            {
                HDC hdcMem;
                HBITMAP hBitmap;
				RECT cr;

                hDC = GetDC(hWnd);
                hdcMem = CreateCompatibleDC(hDC);
				GetClientRect(hWnd, &cr);
                hBitmap = CreateCompatibleBitmap(hDC,
                            cr.right, cr.bottom);
                if (hBitmap) {
                    SelectObject(hdcMem, hBitmap);
                    BitBlt(hdcMem, 0, 0, cr.right, cr.bottom,
                            hDC, 0, 0, SRCCOPY);
                    OpenClipboard(hWnd);
                    EmptyClipboard();
                    SetClipboardData(CF_BITMAP, hBitmap);
                    CloseClipboard();
                    DeleteDC(hdcMem);
                    ReleaseDC(hWnd, hDC);
                }
            }
			return 0;
#endif
	}

    return DefScreenSaverProc(hWnd, Message, wParam, lParam);
}

/* RegisterDialogClasses -- Entry point for registering window
							classes required by configuration dialog box.

	Params:  hWnd -- Handle to window

	Return:  None
*/

BOOL WINAPI RegisterDialogClasses(HANDLE hInst)
{
    return TRUE;
}

/* ScreenSaverConfigureDialog -- Dialog box function for configuration dialog.

	Params:  hWnd -- Handle to window

	Return:  None
*/

DialogueProc(ScreenSaverConfigureDialog)
{
    switch (message) {
        case WM_INITDIALOG:
            GetIniEntries();
            GetIniSettings();
            SendDlgItemMessage(hDlg, IDC_CH_SOUND, BM_SETCHECK, soundGunfire, 0);
			SendDlgItemMessage(hDlg, IDC_CH_IMG_SCREEN, BM_SETCHECK,
				!(bmpShow && (bmpfile[0] != 0)), 0);
			SendDlgItemMessage(hDlg, IDC_CH_IMG_BITMAP, BM_SETCHECK,
				(bmpShow && (bmpfile[0] != 0)), 0);
			SetDlgItemText(hDlg, IDC_CH_BMPFILE, bmpfile);
			EnableWindow(GetDlgItem(hDlg, IDC_CH_IMG_BITMAP), bmpfile[0] != 0);  
            return TRUE;

        case WM_COMMAND:
            switch (wParam) {
                case IDC_CH_BROWSE:
			     	{
			     		OPENFILENAME ofn;
			     		char szString[MAX_PATH];
			
			            memset(&ofn, 0, sizeof(ofn));
						ofn.lStructSize = sizeof(OPENFILENAME);
						ofn.hwndOwner = hDlg;
						ofn.lpstrFilter = rfilter(IDS_BMP_FILE_FILTER);
						ofn.lpstrCustomFilter = NULL;
						strcpy(szString, bmpfile);
						ofn.lpstrFile = (LPSTR) szString;
						ofn.nMaxFile = sizeof(szString);
						ofn.lpstrInitialDir = NULL;
						ofn.lpstrTitle = rstring(IDS_BMP_OPEN_TITLE);
						ofn.Flags = OFN_FILEMUSTEXIST | OFN_HIDEREADONLY;
						if (GetOpenFileName((LPOPENFILENAME) &ofn)) {
							SetDlgItemText(hDlg, IDC_CH_BMPFILE, szString);
							EnableWindow(GetDlgItem(hDlg, IDC_CH_IMG_BITMAP), szString[0] != 0);
						}
					}
					break;

                case IDOK:					// OK
					soundGunfire = IsDlgButtonChecked(hDlg, IDC_CH_SOUND);
					GetDlgItemText(hDlg, IDC_CH_BMPFILE, bmpfile, sizeof bmpfile);
					bmpShow = IsDlgButtonChecked(hDlg, IDC_CH_IMG_BITMAP); 
					
					/* Store all configuration parameters in the registry. */
                    {
						HKEY keyh;
						DWORD dispos;
						TCHAR keyname[80];
						
						strcpy(keyname, "Software\\Fourmilab.ch\\");
						strcat(keyname, szAppName);
						if (RegCreateKeyEx(HKEY_CURRENT_USER,
									keyname,					// Subkey name
									0,							// Reserved
									"",							// Bogus class string
									REG_OPTION_NON_VOLATILE,	// Options
									KEY_ALL_ACCESS,				// Access rights			
									NULL,						// Security attributes
									&keyh,						// Returned handle to the key
									&dispos) == ERROR_SUCCESS) { // Returned disposition
#define regStore(name, value) RegSetValueEx(keyh, name, 0, REG_DWORD, (BYTE *) &value, sizeof(value))
#define regStoreS(name, value) RegSetValueEx(keyh, name, 0, REG_SZ, (BYTE *) value, lstrlen(value) + 1)
								regStore(szGunfire, soundGunfire);  
								regStore(szBmpshow, bmpShow);
								regStoreS(szBmpfile, bmpfile);

								RegCloseKey(keyh);
						}
                   }
#undef regStore
#undef regStoreS
					
                    EndDialog(hDlg, TRUE);
                    return TRUE;

                case IDCANCEL:				// Cancel
                    EndDialog(hDlg, FALSE);
                    return TRUE;
            }
            break;
    }
    return FALSE;
}

/*	GetIniSettings -- Get initial settings from the registry
*/

static void GetIniSettings(void)
{
	HKEY keyh;
	TCHAR keyname[80];
	
	strcpy(keyname, "Software\\Fourmilab.ch\\");
	strcat(keyname, szAppName);

	if (RegOpenKeyEx(HKEY_CURRENT_USER,
			keyname,				// Subkey name
			0,						// Reserved
			KEY_QUERY_VALUE,		// Security access mask
			&keyh) == ERROR_SUCCESS) {
		DWORD vsize;

#define regGet(name, value) vsize = sizeof(value); RegQueryValueEx(keyh, name, NULL, NULL, (BYTE *) &value, &vsize)
#define regGetS(name, value) vsize = sizeof(value); RegQueryValueEx(keyh, name, NULL, NULL, (BYTE *) value, &vsize)

		regGet(szGunfire, soundGunfire);
		regGet(szBmpshow, bmpShow);
		regGetS(szBmpfile, bmpfile);
		
		RegCloseKey(keyh);
#undef regGet
#undef regGetS
	} else {
		soundGunfire = TRUE;
		bmpShow = FALSE;
		bmpfile[0] = '\0';
	}
}

static void GetIniEntries(void)
{
  // Load Common Strings from stringtable...
  LoadString(hMainInstance, idsAppName, szAppName, 40); 
  LoadString(hMainInstance, idsScreenSaver, szScreenSaver, 22);
  LoadString(hMainInstance, idsHelpFile, szHelpFile, MAXFILELEN);
  LoadString(hMainInstance, idsNoHelpMemory, szNoHelpMemory, BUFFLEN);
}
